import argparse
import logging
from configparser import RawConfigParser
from os.path import exists
from typing import Any

from bumpx.helpers import ObjectDict
from bumpx.hooks import HOOKS
from bumpx.version import PARTS, Version

logger = logging.getLogger(__name__)

DEFAULTS: dict[str, Any] = {
    "file": None,
    "regex": r'(__version__|VERSION)\s*=\s*(\'|")(?P<version>.+?)(\'|")',
    "encoding": "utf8",
    "vcs": None,
    "commit": True,
    "tag": True,
    "tag_format": "{version}",
    "tag_annotation": None,
    "push": False,
    "verbose": False,
    "dryrun": False,
    "clean": None,
    "tests": None,
    "skip_tests": False,
    "publish": None,
    "bump_only": False,
    "prepare_only": False,
    "files": [],
    "bump": {
        "unsuffix": True,
        "suffix": None,
        "part": None,
        "message": "Bump version {version}",
    },
    "prepare": {
        "unsuffix": False,
        "suffix": None,
        "part": None,
        "message": "Update to version {version} for next development cycle",
    },
}


class ValidationError(ValueError):
    pass


class BumprConfigParser(RawConfigParser):
    """
    A config parser with optional implicit `bumpr:` prefix on sections.

    Allow better isolation in setup.cfg.
    """

    prefix = "bumpr"

    def candidate_sections(self, section: str) -> list[str]:
        return [section, "{0}:{1}".format(self.prefix, section)]

    def has_section(self, section: str) -> bool:
        sections = self.candidate_sections(section)
        return any(RawConfigParser.has_section(self, section) for section in sections)

    def options(self, section: str) -> list[str]:
        for section in self.candidate_sections(section):
            if RawConfigParser.has_section(self, section):
                return RawConfigParser.options(self, section)

    def has_option(self, section, option):
        sections = self.candidate_sections(section)
        return any(
            RawConfigParser.has_option(self, section, option) for section in sections
        )

    def get(self, section, option, **kwargs):
        for section in self.candidate_sections(section):
            if RawConfigParser.has_option(self, section, option):
                return RawConfigParser.get(self, section, option)

    def getboolean(self, section, option):
        for section in self.candidate_sections(section):
            if RawConfigParser.has_option(self, section, option):
                return RawConfigParser.getboolean(self, section, option)

    def items(self, section: str):
        for section in self.candidate_sections(section):
            if RawConfigParser.has_section(self, section):
                return RawConfigParser.items(self, section)


class Config(ObjectDict):
    def __init__(self, source=None, parsed_args=None):
        super(Config, self).__init__(DEFAULTS)
        for hook in HOOKS:
            self[hook.key] = False

        if source:
            for hook in HOOKS:
                if hook.key in source:
                    self.merge({hook.key: hook.defaults})
            self.merge(source)

        if exists("setup.cfg"):
            self.override_from_config("setup.cfg")

        if parsed_args:
            if "config" in parsed_args and exists(parsed_args.config):
                self.override_from_config(parsed_args.config)
            self.override_from_args(parsed_args)

    def override_from_config(self, filename: str) -> None:
        config = BumprConfigParser()
        config.read_file(open(filename))

        # Common options
        if config.has_section("bumpr"):
            for option in config.options("bumpr"):
                if option in (
                    "tag",
                    "commit",
                    "push",
                    "bump_only",
                    "prepare_only",
                    "skip_tests",
                ):
                    self[option] = config.getboolean("bumpr", option)
                elif option == "files":
                    self.files = [
                        name.strip()
                        for name in config.get("bumpr", "files").split("\n")
                        if name.strip()
                    ]
                else:
                    self[option] = config.get("bumpr", option)

        # Bump and next section
        for section in "bump", "prepare":
            for option in "message", "suffix":
                if config.has_option(section, option):
                    self[section][option] = config.get(section, option)
            if config.has_option(section, "unsuffix"):
                self[section]["unsuffix"] = config.getboolean(section, "unsuffix")

            if config.has_option(section, "part"):
                self[section]["part"] = PARTS[config.get(section, "part").lower()]

        for hook in HOOKS:
            if config.has_section(hook.key):
                if not self.get(hook.key, False):
                    self[hook.key] = hook.defaults
                self[hook.key].update(config.items(hook.key))
            else:
                self[hook.key] = False

    def override_from_args(self, parsed_args):
        for arg in "file", "vcs", "files":
            if arg in parsed_args and getattr(parsed_args, arg) not in (
                None,
                [],
                tuple(),
            ):
                self[arg] = getattr(parsed_args, arg)

        for arg in "verbose", "dryrun":
            if arg in parsed_args and getattr(parsed_args, arg):
                self[arg] = True

        if hasattr(parsed_args, "nocommit"):
            self.commit = not parsed_args.nocommit
        for attr in "bump_only", "prepare_only", "push", "skip_tests":
            if hasattr(parsed_args, attr):
                self[attr] = getattr(parsed_args, attr)

        # Bump
        if parsed_args.part is not None:
            self.bump.part = parsed_args.part
        if parsed_args.suffix is not None:
            self.bump.suffix = parsed_args.suffix
        if parsed_args.unsuffix is not None:
            self.bump.unsuffix = parsed_args.unsuffix

        # Next
        if parsed_args.prepare_part is not None:
            self.prepare.part = parsed_args.prepare_part
        if parsed_args.prepare_suffix is not None:
            self.prepare.suffix = parsed_args.prepare_suffix
        if parsed_args.prepare_unsuffix is not None:
            self.prepare.unsuffix = parsed_args.prepare_unsuffix

    def validate(self):
        if not self.file:
            raise ValidationError(
                "A file is required from the configuration file or the command line"
            )

    @classmethod
    def parse_args(cls, args=None):
        from bumpx import __description__, __version__

        parser = argparse.ArgumentParser(description=__description__)

        parser.add_argument("file", help="Versionned module file", nargs="?")
        parser.add_argument("files", help="Files to update", nargs="*")

        parser.add_argument("--version", action="version", version=__version__)
        parser.add_argument(
            "-v",
            "--verbose",
            dest="verbose",
            action="store_true",
            help="Verbose output",
        )
        parser.add_argument(
            "-c", "--config", default="bumpr.rc", help="Specify a configuration file"
        )
        parser.add_argument(
            "-d",
            "--dryrun",
            action="store_true",
            help="Do not write anything and display a diff",
        )
        parser.add_argument(
            "-st",
            "--skip-tests",
            action="store_true",
            default=argparse.SUPPRESS,
            help="Skip tests",
        )

        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "-b",
            "--bump",
            dest="bump_only",
            action="store_true",
            default=argparse.SUPPRESS,
            help="Only perform the bump",
        )
        group.add_argument(
            "-pr",
            "--prepare",
            dest="prepare_only",
            action="store_true",
            default=argparse.SUPPRESS,
            help="Only perform the prepare",
        )

        # Bump behavior for bump
        group = parser.add_argument_group("bump")
        group.add_argument(
            "-M",
            "--major",
            dest="part",
            action="store_const",
            const=Version.MAJOR,
            help="Bump major version",
        )
        group.add_argument(
            "-m",
            "--minor",
            dest="part",
            action="store_const",
            const=Version.MINOR,
            help="Bump minor version",
        )
        group.add_argument(
            "-p",
            "--patch",
            dest="part",
            action="store_const",
            const=Version.PATCH,
            help="Bump patch version",
        )
        group.add_argument("-s", "--suffix", dest="suffix", type=str, help="Set suffix")
        group.add_argument(
            "-u",
            "--unsuffix",
            dest="unsuffix",
            action="store_true",
            default=None,
            help="Unset suffix",
        )

        # Bump behavior for prepare version
        group = parser.add_argument_group("prepare")
        group.add_argument(
            "-pM",
            "--prepare-major",
            dest="prepare_part",
            action="store_const",
            const=Version.MAJOR,
            help="Bump major version",
        )
        group.add_argument(
            "-pm",
            "--prepare-minor",
            dest="prepare_part",
            action="store_const",
            const=Version.MINOR,
            help="Bump minor version",
        )
        group.add_argument(
            "-pp",
            "--prepare-patch",
            dest="prepare_part",
            action="store_const",
            const=Version.PATCH,
            help="Bump patch version",
        )
        group.add_argument(
            "-ps",
            "--prepare-suffix",
            dest="prepare_suffix",
            type=str,
            help="Set suffix",
        )
        group.add_argument(
            "-pu",
            "--prepare-unsuffix",
            dest="prepare_unsuffix",
            action="store_true",
            help="Unset suffix",
        )

        group = parser.add_argument_group("Version control system")
        group.add_argument(
            "--vcs", choices=["git", "hg"], default=None, help="VCS implementation"
        )
        group.add_argument(
            "-nc",
            "--nocommit",
            action="store_true",
            default=argparse.SUPPRESS,
            help="Do not commit",
        )
        group.add_argument(
            "-P",
            "--push",
            action="store_true",
            default=argparse.SUPPRESS,
            help="Push changes to remote repository",
        )
        group.add_argument(
            "-nP",
            "--no-push",
            action="store_false",
            dest="push",
            default=argparse.SUPPRESS,
            help="Don't push changes to remote repository",
        )

        parsed_args = parser.parse_args(args)
        return cls(parsed_args=parsed_args)



import logging
import shlex
import subprocess

from typing import Any

class BumprError(Exception):
    pass


def check_output(*args, **kwargs):
    """A wrapper for preconfigured calls to subprocess.check_output"""
    return subprocess.check_output(
        stderr=subprocess.STDOUT, universal_newlines=True, *args, **kwargs
    )


def execute(command, verbose: bool=False, replacements=None, dryrun: bool=False):
    logger = logging.getLogger(__name__)
    replacements = replacements or {}
    if not command:
        return
    elif isinstance(command, (list, tuple)):
        if not isinstance(command[0], (list, tuple)):
            command = [command]
        commands = []
        for cmd in command:
            commands.append([part.format(**replacements) for part in cmd])
    else:
        command = command.format(**replacements)
        commands = [
            shlex.split(cmd.strip()) for cmd in command.splitlines() if cmd.strip()
        ]

    output = ""
    for cmd in commands:
        try:
            if dryrun:
                logger.dryrun("execute: {0}".format(" ".join(cmd)))
            elif verbose:
                subprocess.check_call(cmd)
            else:
                output += check_output(cmd)
        except subprocess.CalledProcessError as exception:
            if hasattr(exception, "output") and exception.output:
                print(exception.output)
            cmd = " ".join(cmd) if isinstance(cmd, (list, tuple)) else cmd
            raise BumprError(
                'Command "{0}" failed with exit code {1}'.format(
                    cmd, exception.returncode
                )
            )
    return output


class ObjectDict(dict):
    """A dictionnary with object-like attribute access and depp merge"""

    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def __getattr__(self, key: str):
        return self[key]

    def __setattr__(self, key: str, value: Any) -> None:
        if isinstance(value, dict) and not isinstance(value, ObjectDict):
            value = ObjectDict(value)
        self[key] = value

    def __setitem__(self, key: str, value):
        if isinstance(value, dict) and not isinstance(value, ObjectDict):
            value = ObjectDict(value)
        super(ObjectDict, self).__setitem__(key, value)

    def update(self, *args, **kwargs):
        for key, value in dict(*args, **kwargs).items():
            if isinstance(value, dict) and not isinstance(value, ObjectDict):
                value = ObjectDict(value)
            self[key] = value

    def merge(self, *args, **kwargs):
        for key, value in dict(*args, **kwargs).items():
            if isinstance(value, dict):
                if not isinstance(value, ObjectDict):
                    value = ObjectDict(value)
                if key in self and isinstance(self[key], ObjectDict):
                    self[key].merge(value)
                    continue
            self[key] = value



import logging
from os.path import exists
from typing import TYPE_CHECKING

from .helpers import BumprError, execute

if TYPE_CHECKING:
    from typing import Optional

logger = logging.getLogger(__name__)

__all_ = (
    "Hook",
    "ReadTheDocHook",
    "ChangelogHook",
    "CommandHook",
    "ReplaceHook",
    "HOOKS",
)


class Hook:
    key: str = ""
    defaults: dict[str, Optional[str]] = {}

    def __init__(self, releaser):
        self.releaser = releaser
        self.verbose = releaser.config.verbose
        self.dryrun = releaser.config.dryrun
        self.config = releaser.config[self.key]
        self.validate()

    def validate(self):
        """Override this method to implement initial validation"""

    def bump(self, replacements):
        pass

    def prepare(self, replacements):
        pass


class ReadTheDocHook(Hook):
    """
    This hook set the readthedoc url corresponding to the version
    """

    key = "readthedoc"
    defaults = {
        "id": None,
        "url": "https://{id}.readthedocs.io/en/{tag}",
        "badge": "https://readthedocs.org/projects/{id}/badge/?version={tag}",
        "bump": "{version}",
        "prepare": "latest",
    }

    def url(self, tag):
        return self.config.url.format(id=self.config.id, tag=tag)

    def badge(self, tag):
        return self.config.badge.format(id=self.config.id, tag=tag)

    def bump(self, replacements):
        replacements.insert(
            0, (self.badge("latest"), self.badge(self.releaser.tag_label))
        )
        replacements.insert(0, (self.url("latest"), self.url(self.releaser.tag_label)))

    def prepare(self, replacements):
        replacements.insert(
            0, (self.badge(self.releaser.tag_label), self.badge("latest"))
        )
        replacements.insert(0, (self.url(self.releaser.tag_label), self.url("latest")))


class ChangelogHook(Hook):
    """
    This hook bump the changelog version header and prepare a new section for the next release.
    """

    key = "changelog"
    defaults = {
        "file": None,
        "separator": "-",
        "bump": "{version} ({date:%Y-%m-%d})",
        "prepare": "Current",
        "empty": "Nothing yet",
    }

    def validate(self):
        if not self.config.get("file"):
            raise BumprError("Changelog file has not been specified")
        elif not exists(self.config.file):
            raise BumprError("Changelog file does not exists")

    def bump(self, replacements):
        with open(
            self.config.file, "r", encoding=self.releaser.config.encoding
        ) as changelog_file:
            before = changelog_file.read()
            after = before.replace(self.dev_header(), self.bumped_header())
        self.releaser.perform(self.config.file, before, after)

    def prepare(self, replacements):
        next_header = "\n".join(
            (
                self.dev_header(),
                "",
                "- {0}".format(self.config.empty),
                "",
                self.bumped_header(),
            )
        )
        with open(
            self.config.file, "r", encoding=self.releaser.config.encoding
        ) as changelog_file:
            before = changelog_file.read()
            after = before.replace(self.bumped_header(), next_header)
        self.releaser.perform(self.config.file, before, after)

    def dev_header(self):
        return self.underline(self.config.prepare)

    def bumped_header(self):
        title = self.config.bump.format(
            version=self.releaser.version,
            date=self.releaser.timestamp,
            **self.releaser.version.__dict__,
        )
        return self.underline(title)

    def underline(self, text):
        if self.config.separator:
            return "\n".join((text, len(text) * self.config.separator))
        else:
            return text


class CommandsHook(Hook):
    """
    This hook execute commands
    """

    key = "commands"
    defaults = {
        "bump": None,
        "prepare": None,
    }

    def bump(self, replacements):
        if self.config.bump:
            replacements = dict(
                version=self.releaser.version,
                tag=self.releaser.tag_label,
                date=self.releaser.timestamp,
                **self.releaser.version.__dict__,
            )
            execute(
                self.config.bump,
                replacements=replacements,
                verbose=self.verbose,
                dryrun=self.dryrun,
            )

    def prepare(self, replacements):
        if self.config.prepare:
            replacements = dict(
                version=self.releaser.next_version,
                tag=self.releaser.tag_label,
                date=self.releaser.timestamp,
                **self.releaser.next_version.__dict__,
            )
            execute(
                self.config.prepare,
                replacements=replacements,
                verbose=self.verbose,
                dryrun=self.dryrun,
            )


class ReplaceHook(Hook):
    """
    This hook perform replacements in files
    """

    key = "replace"
    defaults: dict[str, Optional[str]] = {}

    def bump(self, replacements):
        replacements.insert(
            0,
            (
                self.config.dev.format(
                    version=self.releaser.prev_version,
                    tag=self.releaser.tag_label,
                    date=self.releaser.timestamp,
                    **self.releaser.prev_version.__dict__,
                ),
                self.config.stable.format(
                    version=self.releaser.version,
                    tag=self.releaser.tag_label,
                    date=self.releaser.timestamp,
                    **self.releaser.version.__dict__,
                ),
            ),
        )

    def prepare(self, replacements):
        replacements.insert(
            0,
            (
                self.config.stable.format(
                    version=self.releaser.version,
                    tag=self.releaser.tag_label,
                    date=self.releaser.timestamp,
                    **self.releaser.version.__dict__,
                ),
                self.config.dev.format(
                    version=self.releaser.next_version,
                    tag=self.releaser.tag_label,
                    date=self.releaser.timestamp,
                    **self.releaser.next_version.__dict__,
                ),
            ),
        )


HOOKS = (ReadTheDocHook, ChangelogHook, CommandsHook, ReplaceHook)



import logging
from os.path import isdir

from .helpers import BumprError, execute

log = logging.getLogger(__name__)

MSG = "The current repository contains modified files"


class BaseVCS:
    def __init__(self, verbose=False):
        self.verbose = verbose

    def execute(self, command):
        """Execute a command"""
        execute(command, verbose=self.verbose)

    def validate(self, dryrun=False):
        """Ensure the working dir is a repository and there is no modified files"""
        raise NotImplementedError

    def commit(self, message):
        """Commit all modified files"""
        raise NotImplementedError

    def tag(self, name, annotation=None):
        """Create a tag"""
        raise NotImplementedError

    def push(self):
        """Push changes to remote repository"""
        raise NotImplementedError


class Git(BaseVCS):
    def validate(self, dryrun=False):
        if not isdir(".git"):
            raise BumprError("Current directory is not a git repopsitory")

        for line in execute("git status --porcelain", verbose=False).splitlines():
            if not line.startswith("??"):
                if dryrun:
                    log.warning(MSG)
                    break
                else:
                    raise BumprError(MSG)

    def commit(self, message):
        self.execute(["git", "commit", "-am", message])

    def tag(self, name, annotation=None):
        cmd = ["git", "tag", name]
        if annotation:
            cmd += ["--annotate", "-m", '"{0}"'.format(annotation)]
        self.execute(cmd)

    def push(self):
        self.execute(["git", "push"])
        self.execute(["git", "push", "--tags"])


class Mercurial(BaseVCS):
    def validate(self, dryrun=False):
        if not isdir(".hg"):
            raise BumprError("Current directory is not a mercurial repopsitory")

        for line in execute("hg status -mard", verbose=False).splitlines():
            if not line.startswith("??"):
                if dryrun:
                    log.warning(MSG)
                    break
                else:
                    raise BumprError(MSG)

    def commit(self, message):
        self.execute(["hg", "commit", "-A", "-m", message])

    def tag(self, name, annotation=None):
        cmd = ["hg", "tag", name]
        if annotation:
            cmd += ["-m", '"{0}"'.format(annotation)]
        self.execute(cmd)

    def push(self):
        self.execute(["hg", "push"])


class Bazaar(BaseVCS):
    def validate(self, dryrun=False):
        if not isdir(".bzr"):
            raise BumprError("Current directory is not a bazaar repopsitory")

        for line in execute("bzr status --short", verbose=False).splitlines():
            if not line.startswith("?"):
                if dryrun:
                    log.warning(MSG)
                    break
                else:
                    raise BumprError(MSG)

    def commit(self, message):
        self.execute(["bzr", "commit", "-m", message])

    def tag(self, name, annotation=None):
        if annotation:
            log.warning("Tag annotation is not supported by Bazaar")
        self.execute(["bzr", "tag", name])

    def push(self):
        self.execute(["bzr", "push"])


class Fake(BaseVCS):
    def validate(self, dryrun=False):
        return True


VCS = {
    "git": Git,
    "hg": Mercurial,
    "bzr": Bazaar,
    "fake": Fake,
}



import sys
from logging import DEBUG, INFO, getLogger


def main():
    from . import log

    log.init()

    from .config import Config, ValidationError
    from .helpers import BumprError
    from .releaser import Releaser

    config = Config.parse_args()
    getLogger().setLevel(DEBUG if config.verbose else INFO)
    logger = getLogger(__name__)

    try:
        config.validate()
    except ValidationError as e:
        msg = "Invalid configuration: {0}".format(e)
        logger.error(msg)
        sys.exit(1)

    try:
        releaser = Releaser(config)
        releaser.release()
    except BumprError as error:
        logger.error(str(error))
        sys.exit(1)


if __name__ == "__main__":
    main()



"""
Bump'X: Version bumper and Python package releaser

- Clean-up release artifact
- Bump version and tag it
- Build a source distrbution and upload on PyPI
- Update version for new develpoment cycle
- Can run test suite before
- Can be customized with a config file
- Extensible with hooks
"""

from .__about__ import __description__, __version__  # noqa



import logging
import re
from datetime import datetime
from difflib import unified_diff

from .helpers import BumprError, execute
from .hooks import HOOKS
from .vcs import VCS
from .version import Version

logger = logging.getLogger(__name__)


class Releaser:
    """
    Release workflow executor
    """

    def __init__(self, config):
        self.config = config

        with open(config.file) as f:
            match = re.search(config.regex, f.read())
            try:
                version_string = match.group("version")
                self.prev_version = Version.parse(version_string)
            except Exception:
                raise BumprError(
                    "Unable to extract version from {0}".format(config.file)
                )

        logger.debug("Previous version: {0}".format(self.prev_version))

        self.version = self.prev_version.copy()
        self.version.bump(config.bump.part, config.bump.unsuffix, config.bump.suffix)
        logger.debug("Bumped version: {0}".format(self.version))

        self.next_version = self.version.copy()
        self.next_version.bump(
            config.prepare.part, config.prepare.unsuffix, config.prepare.suffix
        )
        logger.debug("Prepared version: {0}".format(self.next_version))

        self.tag_label = self.config.tag_format.format(version=self.version)
        logger.debug("Tag: {0}".format(self.tag_label))
        if self.config.tag_annotation:
            self.tag_annotation = self.config.tag_annotation.format(
                version=self.version
            )
            logger.debug("Tag annotation: {0}".format(self.tag_annotation))

        self.timestamp = None

        if config.vcs:
            self.vcs = VCS[config.vcs](verbose=config.verbose)
            self.vcs.validate(dryrun=config.dryrun)

        if config.dryrun:
            self.modified = {}
            self.diffs = {}

        self.hooks = [hook(self) for hook in HOOKS if self.config[hook.key]]

    def execute(self, command, version=None, verbose=None):
        version = version or self.version
        verbose = verbose or self.config.verbose
        replacements = dict(version=version, date=self.timestamp, **version.__dict__)
        execute(
            command,
            replacements=replacements,
            dryrun=self.config.dryrun,
            verbose=verbose,
        )

    def release(self):
        self.timestamp = datetime.now()

        if self.config.bump_only:
            self.bump()
        elif self.config.prepare_only:
            self.prepare()
        else:
            self.clean()
            self.test()
            self.bump()
            self.publish()
            self.prepare()
            self.push()

    def test(self):
        if self.config.tests:
            if self.config.skip_tests:
                logger.info("Skip test suite")
                return
            logger.info("Running test suite")
            self.execute(self.config.tests, verbose=True)

    def bump(self):
        logger.info(f"Bump version {self.version}")

        replacements = [(str(self.prev_version), str(self.version))]

        for hook in self.hooks:
            hook.bump(replacements)

        self.bump_files(replacements)

        if self.config.vcs:
            self.commit(
                self.config.bump.message.format(
                    version=self.version,
                    tag=self.tag_label,
                    date=self.timestamp,
                    **self.version.__dict__,
                )
            )
            self.tag()

        if self.config.dryrun:
            self.display_diff()
            self.diffs.clear()

    def prepare(self):
        if self.version == self.next_version:
            logger.info("Skip prepare phase")
            return
        logger.info(f"Prepare version {self.next_version}")

        replacements = [(str(self.version), str(self.next_version))]

        for hook in self.hooks:
            hook.prepare(replacements)

        self.bump_files(replacements)

        if self.config.vcs:
            self.commit(
                self.config.prepare.message.format(
                    version=self.next_version,
                    tag=self.tag_label,
                    date=self.timestamp,
                    **self.next_version.__dict__,
                )
            )

        if self.config.dryrun:
            self.display_diff()

    def clean(self):
        """Clean the workspace"""
        if self.config.clean:
            logger.info("Cleaning")
            self.execute(self.config.clean)

    def perform(self, filename, before, after):
        if before == after:
            return
        if self.config.dryrun:
            self.modified[filename] = after
            diff = unified_diff(before.split("\n"), after.split("\n"), lineterm="")
            self.diffs[filename] = diff
        else:
            with open(filename, "w", encoding=self.config.encoding) as f:
                f.write(after)

    def bump_files(self, replacements):
        for filename in [self.config.file] + self.config.files:
            if self.config.dryrun and filename in self.modified:
                before = self.modified[filename]
            else:
                with open(filename, "r", encoding=self.config.encoding) as current_file:
                    before = current_file.read()
            after = before
            for token, replacement in replacements:
                after = after.replace(token, replacement)
            self.perform(filename, before, after)

    def publish(self):
        """Publish the current release to PyPI"""
        if self.config.publish:
            logger.info("Publish")
            self.execute(self.config.publish)

    def tag(self):
        if self.config.commit and self.config.tag:
            if self.config.tag_annotation:
                logger.debug(f"Tag: {self.tag_label} Annotation: {self.tag_annotation}")
                if not self.config.dryrun:
                    self.vcs.tag(self.tag_label, self.tag_annotation)
                else:
                    logger.dryrun(
                        "tag: {0} annotation: {1}".format(
                            self.tag_label, self.tag_annotation
                        )
                    )
            else:
                logger.debug(f"Tag: {self.tag_label}")
                if not self.config.dryrun:
                    self.vcs.tag(self.tag_label)
                else:
                    logger.dryrun("tag: {0}".format(self.tag_label))

    def commit(self, message):
        if self.config.commit:
            logger.debug(f"Commit: {message}")
            if not self.config.dryrun:
                self.vcs.commit(message)
            else:
                logger.dryrun("commit: {0}".format(message))

    def push(self):
        if self.config.vcs and self.config.commit and self.config.push:
            logger.info("Push to upstream repository")
            if not self.config.dryrun:
                self.vcs.push()
            else:
                logger.dryrun("push to remote repository")

    def display_diff(self):
        for filename, diff in self.diffs.items():
            logger.diff(filename)
            for line in diff:
                logger.diff(line)
            logger.diff("")



import logging
import os
import sys
from logging import DEBUG, INFO, Formatter, StreamHandler
from typing import cast

__all__ = ("init",)

RESET_TERM = "\033[0;m"

COLOR_CODES = {
    "red": 31,
    "green": 32,
    "yellow": 33,
    "blue": 34,
    "magenta": 35,
    "cyan": 36,
    "white": 37,
    "bgred": 41,
    "bggrey": 100,
}

LEVEL_COLORS = {
    "DEBUG": "blue",
    "WARNING": "yellow",
    "ERROR": "red",
    "CRITICAL": "bgred",
}

DRYRUN = 25
DIFF = 15


def ansi(color, text):
    """Wrap text in an ansi escape sequence"""
    code = COLOR_CODES[color]
    return "\033[1;{0}m{1}{2}".format(code, text, RESET_TERM)


class ANSIFormatter(Formatter):
    """
    Convert a `logging.LogRecord' object into colored text, using ANSI
    escape sequences.
    """

    def format(self, record):
        msg = record.getMessage()
        if record.levelname == "INFO":
            return ansi("cyan", "-> ") + msg
        elif record.levelname == "DRYRUN":
            return ansi("magenta", "dryrun-> ") + msg
        elif record.levelname == "DIFF":
            if msg.startswith("+"):
                return ansi("green", msg)
            elif msg.startswith("-"):
                return ansi("red", msg)
            else:
                return msg
        else:
            color = LEVEL_COLORS.get(record.levelname, "white")
            return ansi(color, record.levelname.lower()) + ": " + msg


class TextFormatter(Formatter):
    """
    Convert a `logging.LogRecord' object into text.
    """

    def format(self, record):
        if not record.levelname or record.levelname in ("INFO", "DIFF"):
            return record.getMessage()
        elif record.levelname == "DRYRUN":
            return "dryrun-> {0}".format(record.getMessage())
        else:
            return record.levelname.lower() + ": " + record.getMessage()


class BumprLogger(logging.Logger):
    def dryrun(self, *args):
        self.log(DRYRUN, *args)

    def diff(self, *args):
        self.log(DIFF, *args)


def declare():
    logging.addLevelName(DRYRUN, "DRYRUN")
    logging.addLevelName(DIFF, "DIFF")
    logging.setLoggerClass(BumprLogger)


def init(level=INFO):
    declare()

    logger = logging.getLogger()
    handler = StreamHandler()

    if os.isatty(sys.stdout.fileno()) and not sys.platform.startswith("win"):
        fmt = ANSIFormatter()
    else:
        fmt = TextFormatter()
    handler.setFormatter(fmt)
    logger.addHandler(handler)

    if level:
        logger.setLevel(level)


if __name__ == "__main__":  # pragma: no cover
    init(level=DEBUG)

    logger = cast(BumprLogger, logging.getLogger(__name__))
    logger.debug("debug")
    logger.dryrun("dryrun")
    logger.diff("diff")
    logger.diff("+ diff")
    logger.diff("- diff")
    logger.info("info")
    logger.warning("warning")
    logger.error("error")
    logger.critical("critical")
